#!/usr/bin/perl -w
##
# DCPU-16 execution environment - hardware profile initialisation
#
# This module provides the initialisation of the hardware mapped devices
# in the core of the DCPU-16 emulation. The mappings are defined in a
# hardware profile which describes the base address of the devices.
#
# Each device is implemented by a device module in the directory DExecHW.
# An instance of the devices is created and attached to the addresses
# through a DExecHWAdapter. The adapter supplies the offset parameter to
# the device object, allowing multiple registers to be easily implemented
# in the device, despite not being known to the execution environment.
#
# The hardware profiles take the form of lines in a file of the form:
#
#   <address> <device> [ '(' <arguments> ')' ]
#
# <address>   : a 4 character hexadecimal address that the hardware is
#               mapped to
# <device>    : device, as a module in the DExecHW directory.
# <arguments> : comma separated arguments to pass to the device
#
# Comments are prefixed by a '#' character.
#
# Devices may be registered in multiple locations with different arguments.
#
# Each device has a window width, which it defines. This is the number of
# registers it provides and which will be allocated in the core.
#
# FIXME: Make this an IO processor which initialises a bunch of hardware
#        structures in the DASM and assigns symbols.
#

package DASMIO::Hardware;

use DASM;
use DCPU;

use DExecHWAdapter;

use Carp;

## Whether we debug the registrations
my $debug_reg = 0;


##
# Create an object upon which we can work.
#
# @param[in] $dasm   DASM object we're working on
#
# @return new object
sub new
{
    my $proto = shift;
    my $dasm  = shift;
    my $class = ref($proto) || $proto;
    my $self  = {};

    bless $self, $class;
    
    $self->{'dasm'} = $dasm;
    
    # Prefix on all hardware symbols
    $self->{'symbolprefix'} = "";
    
    # Separator for symbols
    $self->{'symbolseparator'} = "_";

    # Base symbol name
    $self->{'symbolbase'} = "_Base";
    
    return $self;
}



##
# Register a region of memory with a device.
#
# @param[in] $self      This hardware object
# @param[in] $address   The address to assign the device at
# @param[in] $device    The device name
# @param[in] $args      arrayref of arguments for the device
sub register
{
    my ($self, $address, $device, $args) = @_;
    my $dasm = $self->{'dasm'};
    
    if (!defined $device || $device eq '')
    {
        confess "No device supplied to Hardware::register";
    }
    
    # Load the device class
    eval {
        require "DExecHW/$device.pm";
    };
    if ($@)
    {
        $self->error( sprintf "Cannot attach device '%s' at %04x: $@",
                              $device,
                              $address);
    }

    # Create the object for the hardware device, with the
    # arguments supplied
    my $object = "DExecHW::$device"->new( @$args );
    
    # Create the initial symbol name
    my $symbol = $self->{'symbolprefix'} . $device;
    $symbol =~ s/::/$symbolseparator/g;

    # Set the base symbol to the address
    # FIXME: This does not cope with multiple instances of a device
    $dasm->setsymbol($symbol . $self->{'symbolbase'}, $address, 'constant');

    # Create one adapter for each offset of the window of this device
    my $window = $object->window();
    for my $offset (0..$window-1)
    {
        my $regaddress = $address + $offset;
        
        # FIXME: Replace with a method
        if (defined $dasm->{'core'}->{$regaddress})
        {
            die "Overlapping data/devices at ".sprintf("%04x",$regaddress);
        }
        
        my $obj = new DExecHWAdapter( $object, $offset);
        $dasm->store($regaddress, $obj);
        
        # Create a symbol for this address.
        my $symname = $obj->symbol();
        if (defined $symname)
        {
            $dasm->setsymbol($symbol . $self->{'symbolseparator'} . $symname,
                             $regaddress,
                             'constant');
        }
    }
    
}


##
# Identify whether we can handle reading or writing a given file
# by its filename.
#
# @param[in]  $self     Our object
# @param[in]  $filename Filename we're accessing
# @param[in]  $type     'read' or 'write'
#
# @retval 1 if we can access it
# @retval 0 if we cannot access it
sub identify
{
    my ($self, $filename, $type) = @_;
    
    if ($filename =~ /\.hw$/ && $type eq 'read')
    {
        return 1;
    }
    
    return 0;
}


##
# List the extensions we support
#
# @param[in]  $self     Our object
#
# @return arrayref of extensions
sub extensions
{
    return [ 'hw' ];
}


##
# Parse an instruction line.
#
# @param[in] $self    IO object
# @param[in] $line    Line to process
sub parseline
{
    my ($self, $line) = @_;
    my $dasm = $self->{'dasm'};

    # Trim any trailing newline that might have been left
    chomp $line;

    if ($line =~ /^\s*#/ ||
        $line =~ /^\s*$/)
    {
        # Skip comments and blank lines
        return;
    }

    # Parse the address and device from:
    #   <hex address> <device name> ...
    my ($address, $device, $_) = ($line =~ /^(?:0x)?([A-Fa-f0-9]+)\s*([\w:]+)\s*(.*$)/);
    if (!defined $address ||
        !defined $device)
    {
        $self->error("Unrecognised line '$_' in hardware profile");
    }

    # Convert the address to a number
    $address = hex($address);

    if ($address >= 0x10000)
    {
        $self->error(sprintf "Address 0x%x too large", $address);
    }

    # The device name can be followed by a list of arguments in brackets,
    # which we parse into an array.
    my $args = [];
    if ($line =~ s/^\(\s*//)
    {
        while (1)
        {
            $line =~ s/\s*//; # Strip leading spaces
            if ($line =~ s/^(\d+)//)
            {
                # Basic number
                push @$args, $1;
            }
            elsif ($line =~ s/^0x([A-Fa-f0-9]+)//)
            {
                # Hex number
                push @$args, hex($1);
            }
            elsif ($line =~ s/^"([^"]+)"// ||
                   $line =~ s/^'([^']+)'//)
            {
                # "d string (double or single)
                push @$args, $1;
            }
            else
            {
                $self->error("Unrecognised argument to $device device: '$line'");
            }

            if ($line =~ s/\s*,\s*//)
            {
                # There's more arguments coming, so go around for
                # more
            }
            elsif ($line =~ s/\s*\)//)
            {
                # That's the end of the arguments, so we can stop
                last;
            }
            elsif ($line =~ /\s*$/)
            {
                # End of the arguments list... odd
                $self->error("Unexpected end of arguments");
            }
            else
            {
                $self->error("Garbage after arguments '$line'");
            }
        }
    }

    # Having parsed the arguments, the rest of the line should be
    # empty (or a comment)
    if (!$line =~ /\s*$/ &&
        !$line =~ /\s*#/)
    {
        $self->error("Garbage at end of device $device: $1");
    }

    printf "Registering at 0x%04x: %s\n", $address, $device if ($debug_reg);
    $self->register($address, $device, $args);
}

# Must return true.
1;
