#!/usr/bin/perl -w
##
# DCPU-16 assembler kit.
#
# based on specification http://0x10c.com/doc/dcpu-16.txt
#
# Syntax: dasm.pl [<options>] <sources>
#
# Example:
#   ./dasm.pl -i <assembler>.dasm -o <output>.dump
#       - Assemble and produce a binary dump
#   ./dasm.pl -i <assembler>.dasm
#       - Assemble and produce an internal structure dump
#   ./dasm.pl -run -i <assembler>.dasm
#       - Assemble and run the program
#   ./dasm.pl -i <output>.dump -o <assembler>.dasm
#       - Disassemble into source
#
#
# Source format is defined in DASMIO::Source.pm.
#
# @file
# @author Justin Fletcher
# 

use Getopt::Long;
use Time::HiRes qw( tv_interval gettimeofday );

use DASM;
use DExec;

## Input filenames
my @inputs;

## Output filename
my $outfile;

## Whether we execute the final code
my $run = 0;

# Create us an assembler data structure
my $dasm = new DASM();

GetOptions("run"        => \$run,
           "input=s"    => \@inputs,
           "outfile=s"  => \$outfile,
           "help"       => sub {
                    print <<EOM;
DASM - DCPU-16 assembler/disassembler kit, by Justin Fletcher.
Syntax: ./dasm.pl [<options>] <input>*

Options:
    -run            Run the program after assembler
    -input <file>   Input file(s)
    -output <file>  Output file
    -help           This help message

File formats and extensions:
EOM
                    for my $io (@{ $dasm->listio() })
                    {
                        printf "    %-12s %s\n",
                               $io->{'name'},
                               (join ", ", map { ".$_" } @{ $io->extensions() });
                    }
                    exit(1);
                }
          ) or exit(1);

# Any remaining arguments are input files
if (@ARGV)
{
    push @inputs, @ARGV;
}

# If they didn't specify inputs, report an error
if (@inputs == 0)
{
    # Help message 
    die "Syntax: $0 [options] <infile>*\n"
}           

# Read in the files (whatever they might be)
for my $infile (@inputs)
{
    $dasm->readfile($infile);
}

if (defined $outfile)
{
    # Write out the file if that's what they requested
    $dasm->writefile($outfile);
    exit(0);
}

if (!$run)
{
    # They didn't supply an output file, so just dump what we have
    $dasm->dump();
    exit(0);
}
else
{
    # FIXME: Although the execute function supports multiple CPUs,
    # there is no way to indicate this on the command line.
    # The intention is that the different CPUs may have different
    # programs and even different hardware in them. There is currently
    # no mechanism to communicate between them - hardware may be
    # implemented by the current hardware interface doesn't know about
    # accessing different CPUs.
    execute($dasm);
}


##
# Execute a, or multiple, programs.
#
# Each DAsm environment will be executed until it stalls. Each
# environment will be run, lowest cycle count first. Runs continue
# until it stalls (pc does not change between calls).
#
# @param[in] @cpus  CPUs to execute
#
# @note returns only when the CPUs have all stalled.
sub execute
{
    my (@cpus) = @_;
    
    for my $dasm (@cpus)
    {
        # Resolve all the relocations
        if (!$dasm->resolve())
        {
            print "Attempting to execute file with unresolved relocations\n";
        }
    }
    
    # Create new execution environments from each core.
    # Each execution environment in the array is a hashref:
    #    'dexec'     => DExec object
    #    'stalled'   => 0 if running, 1 if stalled (pc fixed at an address)
    #    'wallclock' => wallclock time for CPU
    my $cpunum = 0;
    my @execs = map {
        {
            'cpu'       => $cpunum++,
            'dexec'     => new DExec($_),
            'stalled'   => 0,
            'wallclock' => 0,
        }
    } @cpus;
    
    ##
    # List of runnable CPUs
    my @runnable = @cpus;

    # Run until all CPUs have stalled
    while (1)
    {
        # We want to execute the code for the CPU with the least cycles
        # in it.
        my $run = $execs[0];
        my $lowest = $execs[0]->{'dexec'}->{'cycles'};
        for my $exec (@execs[1..$#execs])
        {
            if ($lowest > $exec->{'dexec'}->{'cycles'})
            {
                $run = $exec;
                $lowest = $exec->{'dexec'}->{'cycles'};
            }
        }
        
        my $exec = $run;
        my $dexec = $exec->{'dexec'};

        # Record how long we've been running for on this CPU
        my $starttime = [gettimeofday];

        my $lastpc = $dexec->getreg('pc');

        #$dexec->showregs(); # debugging
        $dexec->runone();

        my $pc = $dexec->getreg('pc');

        # Work out how long we have run for
        my $endtime = [gettimeofday];
        my $wallclock = tv_interval($starttime, $endtime);

        $exec->{'wallclock'} += $wallclock;

        if ($lastpc == $pc)
        {
            # Remove this CPU from the list of those that are runnable
            @runnable = grep {
                $_ != $dexec
            } @runnable;
            
            # Stop if there's nothing left.
            if (scalar(@runnable) == 0)
            {
                last;
            }
        }
    }
    
    # We've finished executing, so lets report some statistics:
    for my $exec (@execs)
    {
        print "\nCPU#$exec->{'cpu'}:\n";
        my $cycles = $exec->{'dexec'}->cycle();

        print "    Executed for $cycles cycles\n";
        printf "    Time taken: %.2f seconds\n", $exec->{'wallclock'};
        printf "    Execution speed: %.2f kHz\n", $exec->{'wallclock'} > 0 ? ($cycles / $exec->{'wallclock'} / 1000) : 0;
    }
}

